mesh_types = {
	"CosN": 0,
	"ConstantN": 1,
	"Cos2N": 2,
	"LinearS": 3,
	"ExponentialS": 4
}

def initHSPM(cfg):
	"""
	Initializes the HSPM object with the given configuration

	Parameters
	----------
	cfg : dict
		Configuration dictionary
		Contains:
			- chord (float): Chord length
			- aoa (float): Angle of attack

			* MESH PARAMETERS
			- meshType (str): Type of mesh (CosN, ConstantN, Cos2N, LinearS, ExponentialS)
			- sLe (float): Leading edge mesh size, only for S-type meshes
			- sTe (float): Trailing edge mesh size, only for S-type meshes
			- naca (str): NACA airfoil code
			or
			- filename (str): Airfoil file names
	"""

	import hspmw
	import numpy as np

	hspm = hspmw.HSPM()

	hspm.chord = cfg['chord']
	hspm.V_inf = 1
	if not 'aoa' in cfg:
		print("HSPM Warning: aoa not defined. Using default value of 0")
		cfg['aoa'] = 0
	hspm.AoA = np.deg2rad(cfg['aoa'])

	if not 'N_wake' in cfg:
		print("HSPM Warning: N_wake not defined. Using default value of 100")
		cfg['N_wake'] = 100
	hspm.N_wake = cfg['N_wake']

	if not 'L_wake' in cfg:
		print("HSPM Warning: L_wake not defined. Using default value of 5")
		cfg['L_wake'] = 5
	hspm.L_wake = cfg['L_wake']

	if not "meshType" in cfg:
		print("HSPM Warning: meshType not defined. Using default cosine mesh (CosN)")
		cfg['meshType'] = "CosN"
	hspm.meshType = mesh_types[cfg['meshType']]
	if cfg['meshType'] == "LinearS" or cfg['meshType'] == "ExponentialS":
		if not "sLe" in cfg:
			print("HSPM Warning: Leading edge mesh size (sLe) not defined. Using default value of 0.01")
			cfg['sLe'] = 0.01
		if not "sTe" in cfg:
			print("HSPM Warning: Trailing edge mesh size (sTe) not defined. Using default value of 0.01")
			cfg['sTe'] = 0.01
		hspm.sLe = cfg['sLe']
		hspm.sTe = cfg['sTe']
		cfg["N"] = 1 # Placeholder value for the generateNaca4DigitCoordinates function
	
	if 'cutTE' in cfg:
		hspm.cutTE = cfg['cutTE']
		if not "sTe" in cfg:
			print("HSPM Warning: Trailing edge mesh size (sTe) not defined. Using default value of 0.01")
			cfg['sTe'] = 0.01
		hspm.sTe = cfg['sTe']

	if 'stopTE' in cfg:
		hspm.stopTE = True 
		hspm.stopTELocation = cfg['stopTE']

	if cfg['naca']:
		hspm.generateNaca4DigitCoordinates(int(cfg['naca'][0]), int(cfg['naca'][1]), int(cfg['naca'][2:4]), int(cfg['N']))
	else:
		hspm.loadCoordinates(cfg['filename'])

	hspm.initHSPM()

	return hspm