#include "geometry.h"
#include <iostream>

void HSPM::generateNaca4DigitCoordinates(double camber, double camberPos, double thickness, size_t _N) 
{
	/*
	Generates the coordinates of a NACA 4 digits airfoil
    Parameters:
    -----------
    camber: double
        Max camber of the airfoil [%]. First digit
    camberPos: double
        Position of the max camber of the airfoil [10%]. Second digit
    thickness: double
        Thickness of the airfoil [%]. Third/fourth digits
	*/

	if (meshType == LinearS) {
		// Figure out how many elements are required
		double N_double = 4.0/(sLe + sTe);
		// Round to the nearest even number
		_N = (size_t) ((int) N_double/2.0) * 2;
		std::cout << "HSPM: The number of elements has been set to " << _N << " to match the prescribed sizes at the LE and TE." << std::endl;
	} else if (meshType == ExponentialS)
	{
		// Figure out how many elements are required
		if (sTe == sLe) { // In this case, we will use constant size elements (might be slightly off the prescribed size, but minimal variation is expected)
			double N_double = 2.0 / sLe;

			// Round to the nearest even number
			_N = (size_t) ((int) N_double/2.0) * 2;
			std::cout << "HSPM: The number of elements has been set to " << _N << " to match the prescribed sizes at the LE and TE." << std::endl;
			
			sLe = 2.0 / _N;
			sTe = 2.0 / _N;

			prog = 1.0;
		} 
		else {
			// Get the progression
			double _r = sTe - sLe + 1;
			double N_double = 2.0 * (log(sTe/sLe) / log(_r) + 1.0);
			// Round to the nearest even number
			_N = (size_t) ((int) N_double/2.0) * 2;
			std::cout << "HSPM: The number of elements has been set to " << _N << " to match the prescribed sizes at the LE and TE." << std::endl;

			// Compute the value of the progression (needs to solve a nonlinear equation w/ Newton's method)
			// 1-r = sLe * (1-r^(N/2))
			auto equation = [this](double r, double _N) {
				return 1-r - sLe * (1-pow(r, _N/2.0));	
			}; 

			auto equationDerivative = [this](double r, double _N) {
				return -1 - sLe * (-pow(r, _N/2.0-1) * _N/2.0);
			};

			double r0 = _r;
			for (int i = 0; i < 1000; ++i) {
				double h = equation(r0, _N) / equationDerivative(r0, _N);
				if (std::abs(h) < 1e-8) {
					break;
				}
				r0 = r0 - h;
			}
			prog = r0;
		}
	}

	N = _N;

	Eigen::VectorXd angleDistribution = Eigen::VectorXd::LinSpaced(N+1, 0, 2*M_PI);
	
	Eigen::VectorXd x_distr = Eigen::VectorXd::Zero(N+1);
	for (size_t i=0; i<=N; i++) {
		double angle = angleDistribution(i);

		if (meshType == CosN) // Cosine mesh with N elements
			x_distr(i) = chord/2 * (1 + cos(angle));
		
		else if (meshType == ConstantN) // Constant size mesh with N elements
		{
			if (angle < M_PI) 
				x_distr(i) = chord * (-M_1_PI * angle + 1);
			else x_distr(i) = chord * (M_1_PI * angle - 1);
		}

		else if (meshType == Cos2N) // Cosine square mesh with N elements
			x_distr(i) = chord * pow((1 + cos((angle + M_PI) / 2)), 2);

		else if (meshType == LinearS) // Linear size mesh, with sizes at LE and TE (approx) prescribed
		{ // Maths au milieu de mon cahier 
			// Compute the size increment (could be saved to memory, but mesh generation is fast enough, so it's fine here)
			double delta_s = (1.0 - (N/2.0)*sLe) / ((N/2.0*(N/2.0-1.0))/2.0);
			if (i <= N/2.0) { // On the top side, from LE to TE
				size_t j = N/2.0-i;

				x_distr(j) = sLe * i + delta_s * (i-1) * i / 2.0;

			} else {
				x_distr(i) = x_distr(i-1) + sLe + delta_s * (i-(N/2.0+1));
			}
		}	

		else if (meshType == ExponentialS) // Exponential size mesh, with sizes @ LE and TE (approx) prescribed
		{
			if (i <= N/2.0) { // Top side
				size_t j = N/2.0-i;

				if (prog == 1) x_distr(j) = sLe * i;
				else x_distr(j) = sLe * (pow(prog, i) - 1) / (prog - 1); 

			} else { // Bottom side
				x_distr(i) = x_distr(i-1) + sLe * pow(prog, i-(N/2.0+1));
			}
		}
	}

	if (cutTE) {
		size_t newN = N;
		size_t remPts = 0; 
		for (size_t i=1; i<=N/2; i++) {
			if (chord - x_distr(i) >= sTe) {
				std::cout << "Removed " << i-1 << " points from the trailing edge. " << std::endl;
				newN = N - (i-1)*2;
				remPts = i-1;
				break;
			}
		}

		Eigen::VectorXd x_distr_new = Eigen::VectorXd(newN+1);
		x_distr_new(0) = x_distr(0);
		for (size_t i=1; i<newN; i++) {
			x_distr_new(i) = x_distr(i+remPts);
		}
		x_distr_new(newN) = x_distr(N);

		N = newN;
		x_distr = x_distr_new;
		std::cout << "HSPM: The mesh contains " << N << " elements." << std::endl;
	}

	if (stopTE) {
		// First, find the mesh size at stopTELOcation
		double stopSize;
		size_t N_behind;
		double actualStopTELocation;
		double behindSize;
		size_t N_rem;
		size_t newN;
		for (size_t i=1; i<N/2; i++) {
			if (x_distr(i) <= stopTELocation) {
				// We found the location
				actualStopTELocation = x_distr(i);
				stopSize = x_distr(i) - x_distr(i+1);
				N_behind = (1 - actualStopTELocation) / stopSize;
				behindSize = (1 - actualStopTELocation) / N_behind;
				N_rem = i; 

				newN = N+2*(N_behind-N_rem);
				break;
			}	
		}

		std::cout << newN << std::endl;

		Eigen::VectorXd x_distr_new = Eigen::VectorXd(newN+1);
		for (size_t i=0; i<N_behind; i++) {
			x_distr_new(i) = 1-behindSize*i;
		}

		for (size_t i=0; i<N-2*N_rem; i++) {
			x_distr_new(i+N_behind) = x_distr(i+N_rem);
		}
		for (size_t i=0; i<N_behind+1; i++) {
			x_distr_new(newN-i) = 1 - i*behindSize;
		}

		N = newN;
		x_distr = x_distr_new;
		std::cout << "HSPM: The mesh contains " << N << " elements." << std::endl;
	}

	x = Eigen::VectorXd(N+1);
	y = Eigen::VectorXd(N+1);

	for (size_t i=0; i<=N; i++) {
		double T = 10 * thickness/100 * chord * (.2969*sqrt(x_distr(i)/chord) - .126*x_distr(i)/chord 
			- .3537*pow(x_distr(i)/chord, 2) + .2843*pow(x_distr(i)/chord, 3) - .1015*pow(x_distr(i)/chord, 4));

		double YBar;
		double dYdX;
		if (x_distr(i) / chord < camberPos/10) {
			YBar = camber/100*x_distr(i)/pow(camberPos/10,2) * (2*camberPos/10 - x_distr(i)/chord);
			dYdX = 2 * camber/100 / pow(camberPos/10,2) * (chord*camberPos/10 - x_distr(i));
		} else {
			YBar = camber/100*(chord-x_distr(i))/pow(1-camberPos/10, 2) * (1+x_distr(i)/chord-2*camberPos/10);
			dYdX = 2 * camber/100 * (chord*camberPos/10 - x_distr(i)) / (chord * pow(1-camberPos/10, 2));
		}

		double X;
		double Y;
		if (i >= N/2) {
			// Lower side
			X = x_distr(i) + T/2 * sin(atan(dYdX));
			Y = YBar - T/2 * cos(atan(dYdX));
		} else {
			// Upper side
			X = x_distr(i) - T/2 * sin(atan(dYdX)); // Could compute the atan only once to speed up, not really useful
													// since this function is only called to create the airfoil
			Y = YBar + T/2 * cos(atan(dYdX));
		}

		x(i) = X;
		y(i) = Y;
	}
}


void HSPM::loadCoordinates(char* fileName) {
    /*
    Loads the coordinates from a file in CSV format.

    Parameters
    ----------
    fileName : char*
        The name of the file to load the coordinates from.
    */

	vector<double> matrixEntries;

	ifstream matrixDataFile(fileName);
	string matrixRowString;
	string matrixEntry;
	int matrixRowNumber = 0;

	while (getline(matrixDataFile, matrixRowString)) {
	    stringstream matrixRowStringStream(matrixRowString);

	    while (getline(matrixRowStringStream, matrixEntry, ' ')) {
	        matrixEntries.push_back(stod(matrixEntry));
	    }
	    matrixRowNumber++;
	}

	Eigen::MatrixXd coordinates = Eigen::Map<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>>(matrixEntries.data(),
            matrixRowNumber, matrixEntries.size() / matrixRowNumber).transpose();
	N = coordinates.cols() - 1;
	x = coordinates.row(0);
	y = coordinates.row(1);

	// TODO: will need to close some airfoils if the first and last points are not the same !!!
}

void HSPM::saveCoordinates(char* fileName) {
    /*
    Saves the coordinates to a file in CSV format.

    Parameters
    ----------
    fileName : char*
        The name of the file to save the coordinates to.
    */
   
	Eigen::MatrixXd coordinates(N+1, 2);
	coordinates.col(0) = x;
	coordinates.col(1) = y;
	saveMatrix(fileName, coordinates);
}