#include "influenceCoeffs.h"

void HSPM::computeConstantInfluenceCoeffs()
{
	/*
	Calculates the constant influence coefficients
	*/

    x_m = Eigen::VectorXd(N);
	y_m = Eigen::VectorXd(N);
	theta = Eigen::VectorXd(N);
	r = Eigen::MatrixXd(N,N+1);
	beta = Eigen::MatrixXd(N,N);

	lengths = Eigen::VectorXd(N);

	for (size_t i=0; i<N; i++) {
		x_m(i) = (x(i) + x(i+1)) / 2.0;
		y_m(i) = (y(i) + y(i+1)) / 2.0;

		theta(i) = atan2(y(i) - y(i+1), x(i) - x(i+1));

		lengths(i) = sqrt(pow(x(i+1) - x(i), 2) + pow(y(i+1) - y(i), 2));

		for (size_t j=0; j<N; j++) {
			beta(i, j) = atan2(y_m(i) - y(j), x_m(i) - x(j)) - atan2(y_m(i) - y(j+1), x_m(i) - x(j+1));
			if (beta(i, j) > M_PI) {
				beta(i, j) -= 2*M_PI;
			} else if (beta(i, j) < -M_PI) {
				beta(i, j) += 2*M_PI;
			}
		}

		for (size_t j=0; j<N+1; j++) {
			r(i, j) = sqrt(pow(x_m(i) - x(j), 2) + pow(y_m(i) - y(j), 2));
		}
	}

	l = lengths.sum();

	A_n = Eigen::MatrixXd(N,N);
	A_t = Eigen::MatrixXd(N,N);

	B_n = Eigen::MatrixXd(N,N);
	B_t = Eigen::MatrixXd(N,N);

    double sine, cosine, logarithm;
	for (size_t i=0; i<N; i++) {
		for (size_t j=0; j<N; j++) {
			if (i == j) {
				A_n(i,j) = 0.5;
				A_t(i,j) = 0;
			} else {
				sine = sin(theta(i) - theta(j));
				cosine = cos(theta(i) - theta(j));
				logarithm = log(r(i,j) / r(i,j+1));
				A_n(i,j) = 1.0/(2*M_PI) * ( sine * logarithm + cosine * beta(i,j) );
				A_t(i,j) = 1.0/(2*M_PI) * ( sine * beta(i,j) - cosine * logarithm );
			}
            
			B_n(i, j) = -A_t(i, j);
			B_t(i, j) = A_n(i, j);
		}
	}
}



Eigen::MatrixXd HSPM::getSpeedAtPoint(double x_p, double y_p) 
{
    /*
    Returns the speed at a given point (x_p, y_p)
    Assumes tau is still unkown at this stage
    The speed is u = a*tau + b, v = c*tau + d
    -------
    Args:
    x_p: double
        x coordinate of the point
    y_p: double
        y coordinate of the point

    Returns:
    Eigen::MatrixXd
        2x2 matrix containing the components (a, b) of speed in x and y directions
    */
    
    Eigen::MatrixXd speeds = Eigen::MatrixXd::Zero(2,2);

    // Freestream
    speeds(0,1) = V_x;
    speeds(1,1) = V_y;

    // Influence of the panels: A and B
    for (size_t j=0; j<N; j++) {
		// Check if we are on the panel
		if (abs(x_p - x_m(j) < 1e-10) && abs(y_p - y_m(j)) < 1e-10)  {
			double _A_t = 0;
			double _A_n = .5;

			double _A_x = _A_t * cos(theta(j)) - _A_n * sin(theta(j));
			double _A_y = _A_t * sin(theta(j)) + _A_n * cos(theta(j));
			double _B_x = _A_y;
			double _B_y = -_A_x;

			// tau contribution
			speeds(0,0) += _A_x * s1(j) + _B_x;
			speeds(1,0) += _A_y * s1(j) + _B_y;

			// constant contribution
			speeds(0,1) += _A_x * s2(j);
			speeds(1,1) += _A_y * s2(j);

			continue;
        }

        double sine = sin(-theta(j));
        double cosine = cos(-theta(j));
        double r_p_jp1 = sqrt( pow(x_p - x(j+1), 2) + pow(y_p - y(j+1), 2) );
        double r_p_j = sqrt( pow(x_p - x(j), 2) + pow(y_p - y(j), 2) );
        double logarithm = log(r_p_j / r_p_jp1);
        double beta_p_jp1 = atan2(y_p - y(j), x_p - x(j)) - atan2(y_p - y(j+1), x_p - x(j+1));
        if (beta_p_jp1 > M_PI) {
            beta_p_jp1 -= 2*M_PI;
        } else if (beta_p_jp1 < -M_PI) {
            beta_p_jp1 += 2*M_PI;
        }

        double _A_x = 1.0/(2*M_PI) * ( sine * beta_p_jp1 - cosine * logarithm );
        double _A_y = 1.0/(2*M_PI) * ( sine * logarithm + cosine * beta_p_jp1 );

        double _B_x = _A_y;
        double _B_y = -_A_x;

        // tau contribution
        speeds(0,0) += _A_x * s1(j) + _B_x;
        speeds(1,0) += _A_y * s1(j) + _B_y;

        // constant contribution
        speeds(0,1) += _A_x * s2(j);
        speeds(1,1) += _A_y * s2(j);
    }

    return speeds;
}