#include "solver.h"

void HSPM::init() 
{
    /*
    Initializes the solver.
    */
    b = Eigen::VectorXd::Zero(N);
    c = Eigen::VectorXd::Zero(N);
}

void HSPM::solve()
{
    /*
    Solves for the steady-state solution.
    */

    this->init();

    // Build the system
    for (size_t i=0; i<N; i++) {
        for (size_t j=0; j<N; j++) {
            b(i) -= B_n(i,j);
        }
        c(i) -=  V_inf * sin(AoA - theta(i));
    }

    c += blVel;

    // Solve using Eigen's solver
    s1 = solver.solve(b);
    s2 = solver.solve(c);

    // q = s1*tau + s2
    // We need Kutta for tau
    tau = this->solveOffBodyKutta();

    q = s1*tau + s2;

    this->computeInviscidVelocity(); 
    this->computePressureDistribution();
}

double HSPM::solveOffBodyKutta() {
    double xc_0 = x_m(0) - dStar(0) * sin(theta(0));
    double yc_0 = y_m(0) + dStar(0) * cos(theta(0));
    double xc_N1 = x_m(N-1) - dStar(N-1) * sin(theta(N-1));
    double yc_N1 = y_m(N-1) + dStar(N-1) * cos(theta(N-1));

    /*
    The magnitude of the velocity at the control points is equal
    We can express any velocity u = a*tau + b
    We have [mag(u,v)]_0^2 - [mag(u,v)]_N1^2 = 0
    or
    [(a*tau + b)^2 + (c*tau + d)^2] - [(e*tau + f)^2 + (g*tau + h)^2] = 0
    developed, this becomes:
    (a^2+c^2-e^2-g^2)*tau^2 + 2*(a*b+c*d-e*f-g*h)*tau + (b^2+d^2-f^2-h^2) = 0
    This can be solved for tau with the quadratic formula
    We just have to compute the coefficients a, b, c, d, e, f, g, h and pick the right solution
    */

    // Left side of the equation
    Eigen::MatrixXd UV_0 = this->getSpeedAtPoint(xc_0, yc_0);
    Eigen::MatrixXd UV_N1 = this->getSpeedAtPoint(xc_N1, yc_N1);

    // Remove the blowing velocity effect
    // Breaks the 15 deg AoA
    /*UV_0(0,1) += blVel(0) * sin(theta(0));
    UV_0(1,1) -= blVel(0) * cos(theta(0));
    UV_N1(0,1) += blVel(N-1) * sin(theta(N-1));
    UV_N1(1,1) -= blVel(N-1) * cos(theta(N-1));*/

    double _a = UV_0(0,0);
    double _b = UV_0(0,1);
    double _c = UV_0(1,0);
    double _d = UV_0(1,1);
    double _e = UV_N1(0,0);
    double _f = UV_N1(0,1);
    double _g = UV_N1(1,0);
    double _h = UV_N1(1,1);

    // Gather all of the terms
    double _tau2_terms = pow(_a,2) + pow(_c,2) - pow(_e,2) - pow(_g,2);
    double _tau_terms = 2*(_a*_b + _c*_d - _e*_f - _g*_h);
    double _const_terms = pow(_b,2) + pow(_d,2) - pow(_f,2) - pow(_h,2);

    if (abs(_tau2_terms) < 1e-10) 
        return -_const_terms / _tau_terms; // If the quadratic equation is degenerate, we can solve it directly

    // Solve the quadratic equation
    double tau_ = (-_tau_terms + sqrt(pow(_tau_terms,2) - 4*_tau2_terms*_const_terms)) / (2*_tau2_terms);

    return tau_;
}


void HSPM::computeInviscidVelocity()
{
    /*
    Computes the inviscid velocity at the viscous boundary
    */

   for (size_t i=0; i<N; i++) {
        double x_visc = x_m(i) - dStar(i) * sin(theta(i));
        double y_visc = y_m(i) + dStar(i) * cos(theta(i));

        Eigen::MatrixXd UV = this->getSpeedAtPoint(x_visc, y_visc);

        U(i) = UV(0,0) * tau + UV(0,1);
        V(i) = UV(1,0) * tau + UV(1,1);

        // Remove the blowing velocity effect
        //U(i) += blVel(i) * sin(theta(i));
        //V(i) -= blVel(i) * cos(theta(i));
    }

    // In the wake
    for (size_t i=0; i<N_wake; i++) {
        Eigen::MatrixXd UV = this->getSpeedAtPoint(x_m_wake(i), y_m_wake(i));
        
        U_wake(i) = UV(0,0) * tau + UV(0,1);
        V_wake(i) = UV(1,0) * tau + UV(1,1);
    }
}

void HSPM::computePressureDistribution() {
    /*
    Computes the pressure distribution on the airfoil
    */

    Eigen::VectorXd V_t(N);
    Cp = Eigen::MatrixXd(N, 4);

    // Compute the tangential speeds 
    for (size_t i=0; i<N; i++) {
        /*V_t(i) = V_inf * cos(AoA - theta(i));
        for (size_t j=0; j<N; j++) {
            V_t(i) += A_t(i, j) * q(j);
            V_t(i) += B_t(i, j) * tau;
        }

        // Compute the pressure coefficients
        Cp(i, 0) = x_m(i);
        Cp(i, 1) = y_m(i);
        Cp(i, 2) = 0;
        Cp(i, 3) = 1 - pow(V_t(i) / V_inf, 2);*/

        
        // Off body Cp, gives better graphs !
        double x_visc = x_m(i) - dStar(i) * sin(theta(i));
        double y_visc = y_m(i) + dStar(i) * cos(theta(i));
        Eigen::MatrixXd UV = this->getSpeedAtPoint(x_visc, y_visc);

        double _U = UV(0,0) * tau + UV(0,1);
        double _V = UV(1,0) * tau + UV(1,1);
        double V_mag = sqrt(pow(_U,2) + pow(_V,2));

        // Compute the pressure coefficients
        Cp(i, 0) = x_m(i);
        Cp(i, 1) = y_m(i);
        Cp(i, 2) = 0;
        Cp(i, 3) = 1 - pow(V_mag / V_inf, 2);
        
    }

    cl = 0;
    cd = 0;
    cm = 0;
    // Compute cd, cl
    for (size_t i=0; i<N; i++) {
        cl -= Cp(i, 3) * cos(theta(i) - AoA) * lengths(i);
        cd += Cp(i, 3) * sin(theta(i) - AoA) * lengths(i);
        cm -= Cp(i, 3) * (x_m(i) - 0.25) * cos(theta(i) - AoA) * lengths(i);
    }
}